"""django_heaven URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
import sys

from django.conf import settings
from django.urls import path

from django_heaven.examples.http import HeavenTestHttpView
from django_heaven.examples.general import HeavenTestAPIView
from django_heaven.examples.redirect import HeavenTestRedirectView
from django_heaven.examples.json import HeavenTestJsonView, HeavenTestJsonProxyView

try:
    from django_heaven.examples.rest_framework import HeavenTestRESTProxyView, HeavenTestRESTView
    has_rest = True
except ImportError:
    has_rest = False

if not settings.DEBUG and 'test' not in sys.argv:
    raise Warning("Do not use django-heaven testing views in production!")


app_name = 'heaven'

urlpatterns = [
    path('example/list/', HeavenTestAPIView.as_view()),

    # http
    path('example/http/', HeavenTestHttpView.as_view(), name='example_http'),

    # json
    path('example/json/', HeavenTestJsonView.as_view(), name='example_json'),
    path('example/json/proxy/', HeavenTestJsonProxyView.as_view(), name='example_json_proxy'),

    # redirect
    path('example/redirect/', HeavenTestRedirectView.as_view(), name='example_redirect'),
]

if has_rest:
    urlpatterns += [
        # rest framework
        path('example/rest/', HeavenTestRESTView.as_view(), name='example_rest'),
        path('example/rest/proxy/', HeavenTestRESTProxyView.as_view(), name='example_rest_proxy'),
    ]



from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import make_password

from services.base import BaseService
from services.decorators import ServiceFunctionDecorator, service_function_for_write


class UserService(BaseService):
    """ That is the service for the User model of your project """
    model = get_user_model()

    def model_create_method(self) -> callable:
        return self.model.create_user

    @service_function_for_write
    @ServiceFunctionDecorator()
    def create_superuser(self, *args, **kwargs):
        """
        Use that to create superuser model. We use monkey-patching which is not a great idea,
        but you will not typically find yourself creating a lot of superusers with UserService() in the production.
        """
        old_model_create = self.model_create_method
        self.model_create_method = lambda: self.model.create_superuser

        result = self.create(*args, **kwargs)
        self.model_create_method = old_model_create
        return result

    @service_function_for_write
    @ServiceFunctionDecorator()
    def set_password(self, **kwargs):
        """ Use that function to set new user password """
        return self.update(
            instance=self._get_argument_from_kwargs(kwargs=kwargs, argument='instance'),
            password=make_password(kwargs.get('password') or kwargs.get('raw_password'))
        )



""" That file contains responses for pure Django JsonResponse """
import json

from django.core.serializers.json import DjangoJSONEncoder
from django.http import JsonResponse

from responses.base import BaseLoggedResponseMixin


class LoggedJsonResponseMixin(BaseLoggedResponseMixin):
    """
    Use that class in order to create a new JsonResponse() with structured data inside. Mind that
    I always use safe=False, since it is not a great idea from my point of view, and heaven must be a safe place.
    """
    def check_response_is_safe(self, data):
        """ Tests that the JsonResponse() is a safe one """
        try:
            if not isinstance(json.loads(data.content), dict):
                raise ValueError(
                    "JsonResponse() must be a safe one. Change your response structure to a dictionary"
                )
        except json.JSONDecodeError:
            raise ValueError(f"Data provided in JsonResponse() cannot be decoded. Data: {data}")
        except (AttributeError, TypeError, ValueError):
            raise ValueError(f"Provide only JsonResponse() objects in LoggedJsonResponseProxyMixin()")

    def _log_json_response(
        self, log_function: callable, data, log_message: str, encoder=DjangoJSONEncoder, **kwargs,
    ):
        """ That function is used not to copy the code that we use in log_info and log_error """
        result_response = log_function(data=data, log_message=log_message, **kwargs)

        if isinstance(data, JsonResponse):
            self.check_response_is_safe(data)
            return data

        return JsonResponse(result_response, encoder=encoder, safe=True,
                            **(kwargs.get('response_kwargs') or {}))

    def log_response_as_info(self, data, log_message: str, encoder=DjangoJSONEncoder, **kwargs):
        return self._log_json_response(
            log_function=super(LoggedJsonResponseMixin, self).log_response_as_error,
            data=data,
            log_message=log_message,
            encoder=encoder,
            **kwargs,
        )

    def log_response_as_error(self, data, log_message: str, encoder=DjangoJSONEncoder, **kwargs):
        return self._log_json_response(
            log_function=super(LoggedJsonResponseMixin, self).log_response_as_error,
            data=data,
            log_message=log_message,
            encoder=encoder,
            **kwargs,
        )


__all__ = [
    'LoggedJsonResponseMixin',
]





class ServiceException(ValueError):
    """
    That is the exception raised by any function from the BaseService()
    when something bad happens because of the normal query, but something in the database.
    """


class ServiceProgrammingException(SyntaxError):
    """
    That is the exception raised when you have a problem with your code, not user related.
    For example, you forgot to provide arguments in model get() function.
    """



from typing import no_type_check

from django.http import HttpResponseRedirect

from responses.base import BaseLoggedResponseMixin


class LoggedRedirectResponseMixin(BaseLoggedResponseMixin):
    """ That class allows you to still log your responses, but call a redirect instead of dict() """

    @no_type_check
    def log_response_as_info(
        self, redirect_to: str, log_message: str, redirect_code: int, kwargs: dict = None,
    ) -> HttpResponseRedirect:
        """
        That function will call normal LoggedResponseMixin info function, but will return redirect instead.
        :params:
            - redirect_to(str): page you want to redirect your user to
            - kwargs(dict): additional arguments for that page
            - redirect_code(int): permanent(308) or temporary(307) redirect
        :rtype: HttpResponseRedirect
        :returns: redirect to another page
        """
        super(LoggedRedirectResponseMixin, self).log_response_as_info(
            data=None, log_message=log_message, status_code=redirect_code, **kwargs,
        )
        return HttpResponseRedirect(redirect_to, **(kwargs.get('response_kwargs') or {}))

    @no_type_check
    def log_response_as_error(
        self, redirect_to: str, log_message: str, redirect_code: int, kwargs: dict = None,
    ) -> HttpResponseRedirect:
        """
        That function will call normal LoggedResponseMixin error function, but will return redirect instead.
        :params:
            - redirect_to(str): page you want to redirect your user to
            - kwargs(dict): additional arguments for that page
            - redirect_code(int): permanent(308) or temporary(307) redirect
        :rtype: HttpResponseRedirect
        :returns: redirect to another page
        """
        super(LoggedRedirectResponseMixin, self).log_response_as_error(
            data=None, log_message=log_message, status_code=redirect_code, **kwargs,
        )
        return HttpResponseRedirect(redirect_to, **(kwargs.get('response_kwargs') or {}))


__all__ = [
    'LoggedRedirectResponseMixin',
]



"""
That is the file that contains all the base services.
Service is a wrapper for the Model. You will not make any direct ORM queries in your projects.
Instead, your code will have different services for various models. Every service will have optimized
ORM queries with logging and custom error handling. That is, you will split your views and serializers
to work with business logic in services.
"""
from django.conf import settings
from django.db.models import Model

from services.exceptions import ServiceException, ServiceProgrammingException
from services.decorators import ServiceFunctionDecorator, service_function_for_write

SERVICES_SETTINGS = settings.DJANGO_HEAVEN['SERVICES']


class BaseService:
    """ The most base service for the django-heaven. Use that instead of direct model calls """
    model: Model = None    # your django ORM model
    read_only: bool = False  # if you only want to read from that model. May be useful for the read-only models
    raise_exception: bool = SERVICES_SETTINGS.get('RAISE_EXCEPTION', True)
    logger_obj = SERVICES_SETTINGS.get('LOGGER_OBJ')

    def __init__(self, objects=None, instance=None):
        class_name = self.__class__.__name__
        self._objects = objects or self.model.objects

        if self.model is None:
            raise ValueError(f"You need to assign model in {class_name}")
        elif not isinstance(self.read_only, bool):
            raise ValueError(f"'write_only' must be a bool() value in {class_name}")

    @property
    def result(self):
        """ Return the result of the service """
        return self._objects

    def service_function_error_handler(self, exc: Exception):
        """
        That function is called whenever we receive an error in any of service functions.
        By default we will return None or will raise an error depending on self.raise_exception.
        If you do not reassign that, we will get global: settings.DJANGO_HEAVEN.SERVICES.RAISE_EXCEPTION
        value.
        """
        if self.raise_exception:
            raise ServiceException(exc)

    @ServiceFunctionDecorator()
    def get(self, *args, **model_fields):
        if not args and not model_fields:
            raise ServiceProgrammingException("You need to provide *args or **kwargs in service get() function")

        return self._objects.get(*args, **model_fields)

    @ServiceFunctionDecorator()
    def filter(self, *args, **model_fields):
        return self._objects.filter(*args, **model_fields)

    @ServiceFunctionDecorator()
    def all(self):
        return self._objects.all()

    @ServiceFunctionDecorator()
    def order_by(self, *args):
        return self._objects.order_by(*args)

    def _get_argument_from_kwargs(self, kwargs: dict, argument: str):
        try:
            return kwargs[argument]
        except KeyError:
            raise ServiceProgrammingException(
                f"You must provide named-only argument '{argument}' in {self.__class__.__name__}",
            )

    def refresh_from_db(self, **kwargs):
        """ Use that in order to refresh your model from the database"""
        instance = self._get_argument_from_kwargs(kwargs=kwargs, argument='instance')
        return instance.refresh_from_db(fields=kwargs.get('fields'), using=kwargs.get('using'))

    @ServiceFunctionDecorator()
    def first(self):
        return self._objects.first()

    @ServiceFunctionDecorator()
    def last(self):
        return self._objects.last()

    @service_function_for_write
    @ServiceFunctionDecorator()
    def update(self, **kwargs):
        """
        Provide named-only argument 'instance' that will act as an instance you want to update.
        That method will automatically use 'update_fields' for the fields that you want to update.

        Arguments you can provide:
            - instance!: the instance that you want to update
            - using: what database you want to use
        """
        instance = self._get_argument_from_kwargs(kwargs=kwargs, argument='instance')
        using_argument: str = kwargs.get('using')

        for field, value in kwargs.items():
            setattr(instance, field, value)

        instance.save(update_fields=kwargs.keys(), force_update=True, using=using_argument)
        return instance

    def model_create_method(self) -> callable:
        """
        That is the function that returns another function that will be used as a creation method
        for your models. For example, in Django User we have create_user() instead of create().
        You will need to reassign your creation function in here.
        """
        return self.model.objects.create

    @service_function_for_write
    @ServiceFunctionDecorator()
    def create(self, *args, **kwargs):
        """Provide arguments to create an instance of your model."""
        instance = self.model_create_method()(*args, **kwargs)
        return instance

    @service_function_for_write
    @ServiceFunctionDecorator()
    def delete(self, **kwargs):
        """ Delete an instance of your model """
        instance = self._get_argument_from_kwargs(kwargs, argument='instance')
        instance.delete()

    def _bulk_operation(self, bulk_function, **kwargs):
        instances = self._get_argument_from_kwargs(kwargs, 'instances')
        return bulk_function([
            self.model(**instance) for instance in instances
        ], **kwargs.get('arguments'))

    @service_function_for_write
    @ServiceFunctionDecorator()
    def bulk_create(self, **kwargs):
        """
        Use that to create a lot of models at once.
        Provide kwargs in that order:
        instances=[
            {"arg" 1},
            {"arg": 2}
        ]
        """
        return self._bulk_operation(bulk_function=self.model.objects.bulk_create, **kwargs,)

    @service_function_for_write
    @ServiceFunctionDecorator
    def bulk_update(self, **kwargs):
        return self._bulk_operation(bulk_function=self.model.objects.bulk_create, **kwargs,)

    def __str__(self):
        return f"{{{self.__class__.__name__} {self.result}}}"

    def __repr__(self):
        return str(self)



from django.http import JsonResponse

from responses.json import LoggedJsonResponseMixin, LoggedJsonResponseProxyMixin
from responses.tests.base import BaseLoggedResponseMixinTest


class LoggedJsonResponseMixinTest(BaseLoggedResponseMixinTest):
    """ That is the tests for the LoggedJsonResponseMixin responses """
    testing_class = LoggedJsonResponseMixin
    is_proxy = False

    info_data = [1, 2, 3]
    error_data = [4, 5, 6]


class LoggedJsonResponseProxyMixinTest(BaseLoggedResponseMixinTest):
    testing_class = LoggedJsonResponseProxyMixin
    is_proxy = True

    info_data = JsonResponse(data={"message": "Hello"})
    error_data = JsonResponse(data={"really": "yes"})

    def test_safe_is_true_only(self):
        try:
            self.response_class.log_response_as_info(data=self.info_data, log_message="Test log message")
            self.response_class.log_response_as_error(data=self.info_data, log_message="Test log message")
        except Exception as exc:
            self.fail(exc)

        with self.assertRaises(ValueError):
            self.response_class.log_response_as_info(
                data=JsonResponse(data=[1, 2, 3], safe=False),
                log_message="Test log message",
            )

        with self.assertRaises(ValueError):
            self.response_class.log_response_as_error(
                data=JsonResponse(data=[1, 2, 3], safe=False),
                log_message="Test log message",
            )

    def test_wrong_data_provided_raises_value_error(self):
        with self.assertRaises(ValueError):
            self.response_class.log_response_as_info(data=[1, 2, 3], log_message="Test log message")

        with self.assertRaises(ValueError):
            self.response_class.log_response_as_error(data=[1, 2, 3], log_message="Test log message")



from logging import getLogger
from unittest.mock import patch

from django.test import TestCase

from django.conf import settings
from responses.base import BaseLoggedResponseMixin, BaseLoggedResponseProxyMixin


class BaseLoggedResponseMixinTest(TestCase):
    """ That is the base class for the responses tests """
    testing_class = None
    is_proxy = False

    info_data = None
    error_data = None

    @classmethod
    def setUpClass(cls):
        super(BaseLoggedResponseMixinTest, cls).setUpClass()
        cls.response_class: BaseLoggedResponseMixin = cls.testing_class()
        cls.response_class.logger_obj = getLogger(settings.TEST_LOGGER_NAME)

    def test_child_of_base_logged_response_mixin(self):
        self.assertIn(BaseLoggedResponseMixin, self.testing_class.mro())

    def test_child_of_base_logged_response_proxy_mixin(self):
        if self.is_proxy:
            self.assertIn(BaseLoggedResponseProxyMixin, self.testing_class.mro())

    def __test_log_response_base(self, log_function: str):
        with patch.object(self.response_class.logger_obj, log_function) as mock_logger:
            log_message = "test log message"
            getattr(self.response_class, f'log_response_as_{log_function}')(
                data=getattr(self, f"{log_function}_data"), log_message=log_message
            )

            mock_logger.assert_called_once_with(log_message)

    def test_log_response_as_info_logging_message(self):
        self.__test_log_response_base('info')

    def test_log_response_as_error_logging_message(self):
        self.__test_log_response_base('error')




""" That file provides responses for the django-rest-framework """
from typing import no_type_check

try:
    from rest_framework.response import Response
except ImportError:
    raise ImportError("You need to install django-rest-framework in order to use rest_framework responses")

from responses.base import BaseLoggedResponseMixin


class LoggedRESTResponseMixin(BaseLoggedResponseMixin):
    """
    That class helps you to not only log your data, but return
    Django rest framework Response() as the formatted result.
    It will create a new Response() object from the data that you supplied,
    and after that it will pass status_code and *args with **kwargs inside of it.
    """

    def _log_rest_response(
        self, log_function: callable, data, log_message: str, status_code: int, **kwargs,
    ):
        result_data = log_function(data=data, log_message=log_message, **kwargs)
        if isinstance(data, Response):
            return data

        return Response(data=result_data, status=status_code, **(kwargs.get('response_kwargs') or {}))

    @no_type_check
    def log_response_as_error(
        self, data, log_message: str, status_code: int, **kwargs,
    ) -> Response:
        return self._log_rest_response(
            log_function=super(LoggedRESTResponseMixin, self).log_response_as_error,
            data=data,
            log_message=log_message,
            status_code=status_code,
            **kwargs,
        )

    @no_type_check
    def log_response_as_info(
        self, data, log_message: str, status_code: int, **kwargs,
    ) -> Response:
        return self._log_rest_response(
            log_function=super(LoggedRESTResponseMixin, self).log_response_as_info,
            data=data,
            log_message=log_message,
            status_code=status_code,
            **kwargs,
        )


__all__ = [
    "LoggedRESTResponseMixin",
]



""" That file provides responses for the Django HttpResponse """
from responses.base import BaseLoggedResponseMixin


class LoggedHttpResponseMixin(BaseLoggedResponseMixin):
    """
    That class helps you to return HttpResponse as a proxy. A lot of people use render, and it
    is complicated and useless to bring such responses to the same structure, however, logging is an
    important thing and we use that class to log the http responses.
    """


__all__ = [
    "LoggedHttpResponseMixin",
]



from functools import wraps

from django.conf import settings
from django.core.exceptions import FieldError

from services.exceptions import ServiceProgrammingException


SERVICES_SETTINGS = settings.DJANGO_HEAVEN['SERVICES']


class ServiceFunctionDecorator:
    def __init__(
        self, force_error_message: bool = SERVICES_SETTINGS.get('FORCE_ERROR_MESSAGE_ARGUMENT', True),
        force_info_message: bool = SERVICES_SETTINGS.get('FORCE_INFO_MESSAGE_ARGUMENT', True),
    ):
        self.force_error_message = force_error_message
        self.force_info_message = force_info_message

    def __logger_argument_check_forced(self, argument_name: str, kwargs):
        """ Checks that the appropriate logger argument is provided if the user set is as forced """
        if getattr(self, f"force{argument_name}", False):
            if kwargs.get(argument_name) is None:
                raise ValueError(f"You must provide {argument_name} argument")
            del kwargs[argument_name]  # We delete it so it does not interfere with other ORM arguments.

    def format_logger_message(self, message: str, resulted_service) -> str:
        """ Use that function in order to format your message. __format__ will be implemented in the future """
        if resulted_service is None:    # if an error happens, then we cannot use the values of None
            return message

        for replace_key, replace_value in (
            ("$service$", resulted_service),
            ("$result$", resulted_service.result),
        ):
            message = message.replace(replace_key, replace_value)

        return message

    def __call__(self, function):
        """
        That decorator is used on every service function. It either runs the functions,
        logs the result and returns it, or logs the result in service_function_error_handler().

        You need to provide named-only arguments called: 'error_message' and 'info_message'.
        These arguments act as the log messages that we will use in error or success scenarios.
        If you don't want to provide an argument, then try tuning
        settings.DJANGO_HEAVEN.SERVICES.FORCE_ERROR_MESSAGE_ARGUMENT argument and the same one
        with INFO instead of ERROR.

        returned_result_in_info is another argument that you may provide. If so, we will
        automatically replace %result% substring in the info message with the result returned from the
        function().
        """

        def service_function_decorator_wrapper(service, *args, **kwargs):
            error_message = kwargs.get('error_message')
            info_message = kwargs.get('info_message')

            self.__logger_argument_check_forced(argument_name='info_message', kwargs=kwargs)
            self.__logger_argument_check_forced(argument_name='error_message', kwargs=kwargs)

            try:
                new_service = service.__class__(objects=function(service, *args, **kwargs))

                if info_message is not None:
                    service.logger_obj.info(self.format_logger_message(info_message, new_service))

                return new_service

            except (FieldError, ServiceProgrammingException) as exc:
                # Field error is related to the wrong arguments, that may be typo,
                # so we don't want to except these as the normal exception
                raise exc
            except Exception as exc:
                service.logger_obj.error(self.format_logger_message(error_message, None))
                return service.service_function_error_handler(exc=exc)

        return service_function_decorator_wrapper


def service_function_for_write(function: callable):
    """ That decorator marks service function that can change the information in the database """

    @wraps(function)
    def service_function_for_write_wrapper(service, *args, **kwargs):
        if service.read_only:
            raise ServiceProgrammingException(f"You are calling write function on read_only service {service}")

        return function(service, *args, **kwargs)

    return service_function_for_write_wrapper


__all__ = [
    'ServiceFunctionDecorator',
    'service_function_for_write',
]



import os
from setuptools import find_packages, setup


with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

requirements = [
    'django',
]

setup(
    name='django-heaven',
    version='0.0.2',
    packages=['responses', 'services'],
    include_package_data=True,
    install_requires=requirements,
    license='MIT License',
    description='django-heaven brings structure and order to your django projects',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/knucklesuganda/django-heaven/',
    author='Andrey Ivanov',
    author_email='python.on.papyrus@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



""" That file contains base classes for the formatted Responses """
from django.conf import settings


RESPONSES_SETTINGS = settings.DJANGO_HEAVEN['RESPONSES']


class BaseLoggedResponseMixin:
    """
    That class helps you bring all your responses to the similar structure and log the results.
    We follow the simple rule: if your data is an instance of any item of
    settings.DJANGO_HEAVEN.RESPONSES.RAW_TYPES, then we convert it to a dictionary, or using
    self.data_conversion_function().
    """
    logger_obj = RESPONSES_SETTINGS['LOGGER_OBJ']
    excluded_conversion_types = None   # optional argument for the proxy responses
    raw_types = None

    def data_conversion_function(self, data, **kwargs):
        """
        That function accepts the raw data from the self._log_response() and uses it
        to convert it to a similar structured responses.
        """
        return {RESPONSES_SETTINGS['DEFAULT_RESPONSE_VERB']: data}

    def _log_response(self, log_function: callable, data, log_message: str, **kwargs):
        """
        That function is used to log the response and return your converted data.
        Do not use it directly, create a wrapper that will work with it internally or
        use our wrappers.
        :params:
            - log_function(callable): any function that will log the response
            - data(str or dict): data that you provide, if it's an item
                of settings.DJANGO_HEAVEN.RESPONSES.RAW_TYPES, then we will convert it
                to a dict() or using self.data_conversion_function(), otherwise we will just
                leave it as it is.
            - log_message(str): the message that you want to put in logs
        :returns: either dict() or list() with logs and message formatting
        """
        log_function(log_message)

        if isinstance(data, (self.raw_types or RESPONSES_SETTINGS['RAW_TYPES'])):
            data = self.data_conversion_function(data, **kwargs)

        return data

    def log_response_as_info(self, data, log_message: str, **kwargs):
        """
        That function helps you to log the response as an informational message.
        Parameters are the same as the are in the _log_response() function
        """
        return self._log_response(
            data=data,
            log_function=self.logger_obj.info,
            log_message=log_message,
            **kwargs,
        )

    def log_response_as_error(self, data, log_message: str, **kwargs):
        """
        That function helps you to log the response as an error message.
        Parameters are the same as the are in the _log_response() function
        """
        return self._log_response(
            data=data,
            log_function=self.logger_obj.error,
            log_message=log_message,
            **kwargs,
        )


__all__ = [
    'BaseLoggedResponseMixin',
]
